@@ -1,42 +1,43 @@
+import base64
 from inspect import getattr_static
 from typing import Final, Any, cast, Optional
+from algosdk.v2client.algod import AlgodClient
 from algosdk.abi import Method
 from pyteal import (
     SubroutineFnWrapper,
     TealInputError,
     Txn,
     MAX_TEAL_VERSION,
-    MethodConfig,
     ABIReturnSubroutine,
     BareCallActions,
     Expr,
     Global,
     OnCompleteAction,
     OptimizeOptions,
-    Reject,
     Router,
     Bytes,
+    Approve,
 )
 
 from beaker.decorators import (
     get_handler_config,
     MethodHints,
+    MethodConfig,
     create,
-    update,
-    delete,
-    opt_in,
-    close_out,
-    clear_state,
 )
+
 from beaker.state import (
     AccountState,
+    AccountStateBlob,
+    ApplicationStateBlob,
     ApplicationState,
-    DynamicAccountStateValue,
+    ReservedAccountStateValue,
     AccountStateValue,
     ApplicationStateValue,
-    DynamicApplicationStateValue,
+    ReservedApplicationStateValue,
 )
 from beaker.errors import BareOverwriteError
+from beaker.precompile import AppPrecompile, LSigPrecompile
 
 
 def get_method_spec(fn) -> Method:
@@ -69,38 +70,72 @@ class Application:
         """Initialize the Application, finding all the custom attributes and initializing the Router"""
         self.teal_version = version
 
-        # Is there a better way to get all the attrs declared in subclasses?
-        self.attrs = {
+        # Get initial list of all attrs declared
+        initial_attrs = {
             m: (getattr(self, m), getattr_static(self, m))
-            for m in list(set(dir(self.__class__)) - set(dir(super())))
+            for m in sorted(list(set(dir(self.__class__)) - set(dir(super()))))
             if not m.startswith("__")
         }
 
+        # Make sure we preserve the ordering of declaration
+        ordering = [
+            m for m in list(vars(self.__class__).keys()) if not m.startswith("__")
+        ]
+        self.attrs = {k: initial_attrs[k] for k in ordering} | initial_attrs
+
+        # Initialize these ahead of time, may not
+        # be set after init if len(precompiles)>0
+        self.approval_program: Optional[str] = None
+        self.clear_program: Optional[str] = None
+
+        self.on_create = None
+        self.on_update = None
+        self.on_delete = None
+        self.on_opt_in = None
+        self.on_close_out = None
+        self.on_clear_state = None
+
         self.hints: dict[str, MethodHints] = {}
         self.bare_externals: dict[str, OnCompleteAction] = {}
-        self.methods: dict[str, ABIReturnSubroutine] = {}
+        self.methods: dict[str, tuple[ABIReturnSubroutine, Optional[MethodConfig]]] = {}
+        self.precompiles: dict[str, AppPrecompile | LSigPrecompile] = {}
 
-        acct_vals: dict[str, AccountStateValue | DynamicAccountStateValue] = {}
-        app_vals: dict[str, ApplicationStateValue | DynamicApplicationStateValue] = {}
+        acct_vals: dict[
+            str, AccountStateValue | ReservedAccountStateValue | AccountStateBlob
+        ] = {}
+        app_vals: dict[
+            str,
+            ApplicationStateValue
+            | ReservedApplicationStateValue
+            | ApplicationStateBlob,
+        ] = {}
 
-        methods: dict[str, tuple[ABIReturnSubroutine, Optional[MethodConfig]]] = {}
         for name, (bound_attr, static_attr) in self.attrs.items():
-
             # Check for state vals
             match bound_attr:
+
                 case AccountStateValue():
                     if bound_attr.key is None:
                         bound_attr.key = Bytes(name)
                     acct_vals[name] = bound_attr
-                case DynamicAccountStateValue():
+                case ReservedAccountStateValue():
                     acct_vals[name] = bound_attr
+                case AccountStateBlob():
+                    acct_vals[name] = bound_attr
+
+                case ApplicationStateBlob():
+                    app_vals[name] = bound_attr
                 case ApplicationStateValue():
                     if bound_attr.key is None:
                         bound_attr.key = Bytes(name)
                     app_vals[name] = bound_attr
-                case DynamicApplicationStateValue():
+                case ReservedApplicationStateValue():
                     app_vals[name] = bound_attr
 
+                case LSigPrecompile() | AppPrecompile():
+                    self.precompiles[name] = bound_attr
+
+            # Already dealt with these, move on
             if name in app_vals or name in acct_vals:
                 continue
 
@@ -137,10 +172,43 @@ class Application:
                 # Create the ABIReturnSubroutine from the static attr
                 # but override the implementation with the bound version
                 abi_meth = ABIReturnSubroutine(static_attr)
+
                 if handler_config.referenced_self:
                     abi_meth.subroutine.implementation = bound_attr
-                methods[name] = (abi_meth, handler_config.method_config)
 
+                self.methods[name] = (abi_meth, handler_config.method_config)
+
+                if handler_config.is_create():
+                    if self.on_create is not None:
+                        raise TealInputError("Multiple create methods specified")
+                    self.on_create = static_attr
+
+                if handler_config.is_update():
+                    if self.on_update is not None:
+                        raise TealInputError("Multiple update methods specified")
+                    self.on_update = static_attr
+
+                if handler_config.is_delete():
+                    if self.on_delete is not None:
+                        raise TealInputError("Multiple delete methods specified")
+                    self.on_delete = static_attr
+
+                if handler_config.is_opt_in():
+                    if self.on_opt_in is not None:
+                        raise TealInputError("Multiple opt in methods specified")
+                    self.on_opt_in = static_attr
+
+                if handler_config.is_clear_state():
+                    if self.on_clear_state is not None:
+                        raise TealInputError("Multiple clear state methods specified")
+                    self.on_clear_state = static_attr
+
+                if handler_config.is_close_out():
+                    if self.on_close_out is not None:
+                        raise TealInputError("Multiple close out methods specified")
+                    self.on_close_out = static_attr
+
+                self.methods[name] = (abi_meth, handler_config.method_config)
                 self.hints[name] = handler_config.hints()
 
             # Internal subroutines
@@ -157,7 +225,15 @@ class Application:
         self.acct_state = AccountState(acct_vals)
         self.app_state = ApplicationState(app_vals)
 
-        # Create router with name of class and bare externals
+        # If there are no precompiles, we can build the programs
+        # with what we already have
+        if len(self.precompiles) == 0:
+            self.compile()
+
+    def compile(self) -> tuple[str, str]:
+        if self.approval_program is not None and self.clear_program is not None:
+            return self.approval_program, self.clear_program
+
         self.router = Router(
             name=self.__class__.__name__,
             bare_calls=BareCallActions(**self.bare_externals),
@@ -165,13 +241,13 @@ class Application:
         )
 
         # Add method externals
-        for method_name, method_tuple in methods.items():
+        for _, method_tuple in self.methods.items():
             method, method_config = method_tuple
-            self.methods[method_name] = method
             self.router.add_method_handler(
                 method_call=method, method_config=method_config
             )
 
+        # Compile approval and clear programs
         (
             self.approval_program,
             self.clear_program,
@@ -182,10 +258,24 @@ class Application:
             optimize=OptimizeOptions(scratch_slots=True),
         )
 
+        return (self.approval_program, self.clear_program)
+
     def application_spec(self) -> dict[str, Any]:
         """returns a dictionary, helpful to provide to callers with information about the application specification"""
+
+        if self.approval_program is None or self.clear_program is None:
+            raise Exception(
+                "approval or clear program are none, please build the programs first"
+            )
+
         return {
-            "hints": {k: v.dictify() for k, v in self.hints.items()},
+            "hints": {k: v.dictify() for k, v in self.hints.items() if not v.empty()},
+            "source": {
+                "approval": base64.b64encode(self.approval_program.encode()).decode(
+                    "utf8"
+                ),
+                "clear": base64.b64encode(self.clear_program.encode()).decode("utf8"),
+            },
             "schema": {
                 "local": self.acct_state.dictify(),
                 "global": self.app_state.dictify(),
@@ -194,7 +284,12 @@ class Application:
         }
 
     def initialize_application_state(self) -> Expr:
-        """Initialize any application state variables declared"""
+        """
+        Initialize any application state variables declared
+
+        :return: The Expr to initialize the application state.
+        :rtype: pyteal.Expr
+        """
         return self.app_state.initialize()
 
     def initialize_account_state(self, addr=Txn.sender()) -> Expr:
@@ -210,30 +305,38 @@ class Application:
 
     @create
     def create(self) -> Expr:
-        """default create behavior, initializes application state"""
-        return self.initialize_application_state()
+        return Approve()
 
-    @opt_in
-    def opt_in(self) -> Expr:
-        """default opt in behavior, initializes account state"""
-        return self.initialize_account_state()
+    def dump(self, directory: str = ".", client: Optional[AlgodClient] = None):
+        if self.approval_program is None:
+            if client is not None:
+                ac = AppPrecompile(self)
+                ac.compile(client)
+            else:
+                raise Exception(
+                    "Approval program empty, if you have precompiles, pass an Algod client to build the precompiles"
+                )
 
-    @update
-    def update(self) -> Expr:
-        """default update behavior, rejects"""
-        return Reject()
+        import json
+        import os
 
-    @delete
-    def delete(self) -> Expr:
-        """default delete behavior, rejects"""
-        return Reject()
+        if not os.path.exists(directory):
+            os.mkdir(directory)
 
-    @close_out
-    def close_out(self) -> Expr:
-        """default close out behavior, rejects"""
-        return Reject()
+        with open(os.path.join(directory, "approval.teal"), "w") as f:
+            if self.approval_program is None:
+                raise Exception("Approval program empty")
+            f.write(self.approval_program)
 
-    @clear_state
-    def clear_state(self) -> Expr:
-        """default clear state behavior, rejects"""
-        return Reject()
+        with open(os.path.join(directory, "clear.teal"), "w") as f:
+            if self.clear_program is None:
+                raise Exception("Clear program empty")
+            f.write(self.clear_program)
+
+        with open(os.path.join(directory, "contract.json"), "w") as f:
+            if self.contract is None:
+                raise Exception("Contract empty")
+            f.write(json.dumps(self.contract.dictify(), indent=4))
+
+        with open(os.path.join(directory, "application.json"), "w") as f:
+            f.write(json.dumps(self.application_spec(), indent=4))
