@@ -1,13 +1,14 @@
 import pytest
 from typing import Final, cast
+from dataclasses import asdict
 from Cryptodome.Hash import SHA512
 import pyteal as pt
 
 from beaker.state import (
-    DynamicApplicationStateValue,
+    ReservedApplicationStateValue,
     ApplicationStateValue,
     AccountStateValue,
-    DynamicAccountStateValue,
+    ReservedAccountStateValue,
 )
 
 from beaker.errors import BareOverwriteError
@@ -18,15 +19,17 @@ from beaker.application import (
     get_method_spec,
 )
 from beaker.decorators import (
-    ResolvableArguments,
-    ResolvableTypes,
+    DefaultArgumentClass,
     external,
+    get_handler_config,
     internal,
     create,
+    opt_in,
+    clear_state,
+    close_out,
     update,
     delete,
 )
-from beaker.struct import Struct
 
 options = pt.CompileOptions(mode=pt.Mode.Application, version=pt.MAX_TEAL_VERSION)
 
@@ -47,8 +50,8 @@ def test_empty_application():
     ), "Expected no schema"
 
     assert len(ea.bare_externals.keys()) == len(
-        EXPECTED_BARE_externalS
-    ), f"Expected {len(EXPECTED_BARE_externalS)} bare externals: {EXPECTED_BARE_externalS}"
+        EXPECTED_BARE_HANDLERS
+    ), f"Expected {len(EXPECTED_BARE_HANDLERS)} bare handlers: {EXPECTED_BARE_HANDLERS}"
     assert (
         len(ea.approval_program) > 0
     ), "Expected approval program to be compiled to teal"
@@ -83,8 +86,8 @@ def test_single_external():
         sh.contract.get_method_by_name("made up")
 
 
-def test_bare_external():
-    class Bareexternal(Application):
+def test_bare():
+    class Bare(Application):
         @create
         def create():
             return pt.Approve()
@@ -97,18 +100,109 @@ def test_bare_external():
         def delete():
             return pt.Approve()
 
-    bh = Bareexternal()
-    assert len(bh.bare_externals) == len(
-        EXPECTED_BARE_externalS
-    ), f"Expected {len(EXPECTED_BARE_externalS)} bare externals: {EXPECTED_BARE_externalS}"
+    bh = Bare()
 
-    class FailBareexternal(Application):
+    assert (
+        len(bh.bare_externals) == 3
+    ), "Expected 3 bare externals: create,update,delete"
+
+    class FailBare(Application):
         @create
         def wrong_name():
             return pt.Approve()
 
     with pytest.raises(BareOverwriteError):
-        bh = FailBareexternal()
+        bh = FailBare()
+
+
+def test_mixed_bares():
+    class MixedBare(Application):
+        @create
+        def create(self):
+            return pt.Approve()
+
+        @opt_in
+        def opt_in(self, s: pt.abi.String):
+            return pt.Assert(pt.Len(s.get()))
+
+    mb = MixedBare()
+    assert len(mb.bare_externals) == 1
+    assert len(mb.methods) == 1
+
+
+def test_bare_external():
+    class BareExternal(Application):
+        @create
+        def create(self, s: pt.abi.String):
+            return pt.Assert(pt.Len(s.get()))
+
+        @opt_in
+        def opt_in(self, s: pt.abi.String):
+            return pt.Assert(pt.Len(s.get()))
+
+        @close_out
+        def close_out(self, s: pt.abi.String):
+            return pt.Assert(pt.Len(s.get()))
+
+        @clear_state
+        def clear_state(self, s: pt.abi.String):
+            return pt.Assert(pt.Len(s.get()))
+
+        @update
+        def update(self, s: pt.abi.String):
+            return pt.Assert(pt.Len(s.get()))
+
+        @delete
+        def delete(self, s: pt.abi.String):
+            return pt.Assert(pt.Len(s.get()))
+
+    be = BareExternal()
+    assert len(be.bare_externals) == 0, "Should have no bare externals"
+    assert (
+        len(be.contract.methods) == 6
+    ), "should have create, optin, closeout, clearstate, update, delete"
+
+    hc = get_handler_config(BareExternal.create)
+    assert hc.method_config is not None
+    confs = asdict(hc.method_config)
+    assert confs["no_op"] == pt.CallConfig.CREATE
+    del confs["no_op"]
+    assert all([c == pt.CallConfig.NEVER for c in confs.values()])
+
+    hc = get_handler_config(BareExternal.opt_in)
+    assert hc.method_config is not None
+    confs = asdict(hc.method_config)
+    assert confs["opt_in"] == pt.CallConfig.CALL
+    del confs["opt_in"]
+    assert all([c == pt.CallConfig.NEVER for c in confs.values()])
+
+    hc = get_handler_config(BareExternal.close_out)
+    assert hc.method_config is not None
+    confs = asdict(hc.method_config)
+    assert confs["close_out"] == pt.CallConfig.CALL
+    del confs["close_out"]
+    assert all([c == pt.CallConfig.NEVER for c in confs.values()])
+
+    hc = get_handler_config(BareExternal.clear_state)
+    assert hc.method_config is not None
+    confs = asdict(hc.method_config)
+    assert confs["clear_state"] == pt.CallConfig.CALL
+    del confs["clear_state"]
+    assert all([c == pt.CallConfig.NEVER for c in confs.values()])
+
+    hc = get_handler_config(BareExternal.update)
+    assert hc.method_config is not None
+    confs = asdict(hc.method_config)
+    assert confs["update_application"] == pt.CallConfig.CALL
+    del confs["update_application"]
+    assert all([c == pt.CallConfig.NEVER for c in confs.values()])
+
+    hc = get_handler_config(BareExternal.delete)
+    assert hc.method_config is not None
+    confs = asdict(hc.method_config)
+    assert confs["delete_application"] == pt.CallConfig.CALL
+    del confs["delete_application"]
+    assert all([c == pt.CallConfig.NEVER for c in confs.values()])
 
 
 def test_subclass_application():
@@ -152,15 +246,15 @@ def test_app_state():
     assert app.app_state.num_uints == 1, "Expected 1 int"
     assert app.app_state.num_byte_slices == 1, "Expected 1 byte slice"
 
-    class DynamicAppState(BasicAppState):
+    class ReservedAppState(BasicAppState):
         uint_dynamic: Final[
-            DynamicApplicationStateValue
-        ] = DynamicApplicationStateValue(stack_type=pt.TealType.uint64, max_keys=10)
+            ReservedApplicationStateValue
+        ] = ReservedApplicationStateValue(stack_type=pt.TealType.uint64, max_keys=10)
         byte_dynamic: Final[
-            DynamicApplicationStateValue
-        ] = DynamicApplicationStateValue(stack_type=pt.TealType.bytes, max_keys=10)
+            ReservedApplicationStateValue
+        ] = ReservedApplicationStateValue(stack_type=pt.TealType.bytes, max_keys=10)
 
-    app = DynamicAppState()
+    app = ReservedAppState()
     assert app.app_state.num_uints == 11, "Expected 11 ints"
     assert app.app_state.num_byte_slices == 11, "Expected 11 byte slices"
 
@@ -179,15 +273,15 @@ def test_acct_state():
     assert app.acct_state.num_uints == 1, "Expected 1 int"
     assert app.acct_state.num_byte_slices == 1, "Expected 1 byte slice"
 
-    class DynamicAcctState(BasicAcctState):
-        uint_dynamic: Final[DynamicAccountStateValue] = DynamicAccountStateValue(
+    class ReservedAcctState(BasicAcctState):
+        uint_dynamic: Final[ReservedAccountStateValue] = ReservedAccountStateValue(
             stack_type=pt.TealType.uint64, max_keys=5
         )
-        byte_dynamic: Final[DynamicAccountStateValue] = DynamicAccountStateValue(
+        byte_dynamic: Final[ReservedAccountStateValue] = ReservedAccountStateValue(
             stack_type=pt.TealType.bytes, max_keys=5
         )
 
-    app = DynamicAcctState()
+    app = ReservedAcctState()
     assert app.acct_state.num_uints == 6, "Expected 6 ints"
     assert app.acct_state.num_byte_slices == 6, "Expected 6 byte slices"
 
@@ -264,48 +358,188 @@ def test_internal():
         assert actual == expected
 
 
-def test_resolvable_hint():
+def test_default_param_state():
     class Hinty(Application):
-        @external(read_only=True)
-        def get_asset_id(*, output: pt.abi.Uint64):
-            return output.set(pt.Int(123))
+        asset_id = ApplicationStateValue(pt.TealType.uint64, default=pt.Int(123))
 
-        @external(resolvable=ResolvableArguments(aid=get_asset_id))
-        def hintymeth(aid: pt.abi.Asset):
-            return pt.Assert(pt.Int(1))
+        @external
+        def hintymeth(self, num: pt.abi.Uint64, aid: pt.abi.Asset = asset_id):
+            return pt.Assert(aid.asset_id() == self.asset_id)
 
     h = Hinty()
 
     assert h.hintymeth.__name__ in h.hints, "Expected a hint available for the method"
 
     hint = h.hints[h.hintymeth.__name__]
+
+    assert "aid" in hint.default_arguments, "Expected annotation available for param"
+
+    default = hint.default_arguments["aid"]
+
+    assert default.resolvable_class == DefaultArgumentClass.GlobalState
     assert (
-        hint.resolvable["aid"][ResolvableTypes.ABIMethod]
-        == get_method_spec(h.get_asset_id).dictify()
+        default.resolve_hint() == Hinty.asset_id.str_key()
     ), "Expected the hint to match the method spec"
 
 
-EXPECTED_BARE_externalS = [
+def test_default_param_const():
+    const_val = 123
+
+    class Hinty(Application):
+        @external
+        def hintymeth(self, num: pt.abi.Uint64, aid: pt.abi.Asset = const_val):
+            return pt.Assert(aid.asset_id() == pt.Int(const_val))
+
+    h = Hinty()
+
+    assert h.hintymeth.__name__ in h.hints, "Expected a hint available for the method"
+
+    hint = h.hints[h.hintymeth.__name__]
+
+    assert "aid" in hint.default_arguments, "Expected annotation available for param"
+
+    default = hint.default_arguments["aid"]
+
+    assert default.resolvable_class == DefaultArgumentClass.Constant
+    assert (
+        default.resolve_hint() == const_val
+    ), "Expected the hint to match the method spec"
+
+
+def test_default_read_only_method():
+    const_val = 123
+
+    class Hinty(Application):
+        @external(read_only=True)
+        def get_asset_id(self, *, output: pt.abi.Uint64):
+            return output.set(pt.Int(const_val))
+
+        @external
+        def hintymeth(self, num: pt.abi.Uint64, aid: pt.abi.Asset = get_asset_id):
+            return pt.Assert(aid.asset_id() == pt.Int(const_val))
+
+    h = Hinty()
+
+    assert h.hintymeth.__name__ in h.hints, "Expected a hint available for the method"
+
+    hint = h.hints[h.hintymeth.__name__]
+
+    assert "aid" in hint.default_arguments, "Expected annotation available for param"
+
+    default = hint.default_arguments["aid"]
+
+    assert default.resolvable_class == DefaultArgumentClass.ABIMethod
+    assert (
+        default.resolve_hint() == get_method_spec(Hinty.get_asset_id).dictify()
+    ), "Expected the hint to match the method spec"
+
+
+def test_app_spec():
+    class Specd(Application):
+        decl_app_val = ApplicationStateValue(pt.TealType.uint64)
+        decl_acct_val = AccountStateValue(pt.TealType.uint64)
+
+        @external(read_only=True)
+        def get_asset_id(self, *, output: pt.abi.Uint64):
+            return output.set(pt.Int(123))
+
+        @external
+        def annotated_meth(self, aid: pt.abi.Asset = get_asset_id):
+            return pt.Assert(pt.Int(1))
+
+        class Thing(pt.abi.NamedTuple):
+            a: pt.abi.Field[pt.abi.Uint64]
+            b: pt.abi.Field[pt.abi.Uint32]
+
+        @external
+        def struct_meth(self, thing: Thing):
+            return pt.Approve()
+
+    s = Specd()
+
+    actual_spec = s.application_spec()
+
+    get_asset_id_hints = {"read_only": True}
+    annotated_meth_hints = {
+        "default_arguments": {
+            "aid": {
+                "source": "abi-method",
+                "data": {
+                    "name": "get_asset_id",
+                    "args": [],
+                    "returns": {"type": "uint64"},
+                },
+            },
+        }
+    }
+    struct_meth_hints = {
+        "structs": {
+            "thing": {"name": "Thing", "elements": [("a", "uint64"), ("b", "uint32")]}
+        }
+    }
+
+    expected_hints = {
+        "get_asset_id": get_asset_id_hints,
+        "annotated_meth": annotated_meth_hints,
+        "struct_meth": struct_meth_hints,
+    }
+
+    expected_schema = {
+        "local": {
+            "declared": {
+                "decl_acct_val": {
+                    "type": "uint64",
+                    "key": "decl_acct_val",
+                    "descr": "",
+                }
+            },
+            "reserved": {},
+        },
+        "global": {
+            "declared": {
+                "decl_app_val": {
+                    "type": "uint64",
+                    "key": "decl_app_val",
+                    "descr": "",
+                }
+            },
+            "reserved": {},
+        },
+    }
+
+    def dict_match(a: dict, e: dict) -> bool:
+        for k, v in a.items():
+            if type(v) is dict:
+                if not dict_match(v, e[k]):
+                    print(f"comparing {k} {v} {e[k]}")
+                    return False
+            else:
+                if v != e[k]:
+                    print(f"comparing {k}")
+                    return False
+
+        return True
+
+    assert dict_match(actual_spec["hints"], expected_hints)
+    assert dict_match(actual_spec["schema"], expected_schema)
+
+
+EXPECTED_BARE_HANDLERS = [
     "create",
-    "update",
-    "delete",
-    "opt_in",
-    "clear_state",
-    "close_out",
 ]
 
 
-def test_model_args():
+def test_struct_args():
     from algosdk.abi import Method, Argument, Returns
 
     class Structed(Application):
-        class UserRecord(Struct):
-            addr: pt.abi.Address
-            balance: pt.abi.Uint64
-            nickname: pt.abi.String
+        class UserRecord(pt.abi.NamedTuple):
+            addr: pt.abi.Field[pt.abi.Address]
+            balance: pt.abi.Field[pt.abi.Uint64]
+            nickname: pt.abi.Field[pt.abi.String]
 
         @external
-        def structy(user_record: UserRecord):
+        def structy(self, user_record: UserRecord):
             return pt.Assert(pt.Int(1))
 
     m = Structed()
@@ -317,7 +551,11 @@ def test_model_args():
     assert m.hints["structy"].structs == {
         "user_record": {
             "name": "UserRecord",
-            "elements": ["addr", "balance", "nickname"],
+            "elements": [
+                ("addr", "address"),
+                ("balance", "uint64"),
+                ("nickname", "string"),
+            ],
         }
     }
 
