@@ -1,7 +1,8 @@
 from base64 import b64decode
 import copy
+from dataclasses import dataclass
 from math import ceil
-from typing import Any
+from typing import Any, cast, Optional
 
 from algosdk.account import address_from_private_key
 from algosdk.atomic_transaction_composer import (
@@ -30,6 +31,7 @@ from beaker.decorators import (
 )
 from beaker.client.state_decode import decode_state
 from beaker.client.logic_error import LogicException
+from beaker.precompile import AppPrecompile
 
 
 class ApplicationClient:
@@ -52,29 +54,56 @@ class ApplicationClient:
         if signer is not None and sender is None:
             self.sender = self.get_sender(sender, self.signer)
 
+        self.approval_binary: Optional[bytes] = None
+        self.approval_src_map: Optional[SourceMap] = None
+        self.approval_asserts: dict[int, ProgramAssertion] = {}
+
+        self.clear_binary: Optional[bytes] = None
+        self.clear_src_map: Optional[SourceMap] = None
+        self.clear_asserts: dict[int, ProgramAssertion] = {}
+
         self.suggested_params = suggested_params
 
-    def compile_approval(self, source_map: bool = False) -> tuple[bytes, SourceMap]:
-        approval_result = self.client.compile(
-            self.app.approval_program, source_map=source_map
-        )
+    def compile(
+        self, teal: str, source_map: bool = False
+    ) -> tuple[bytes, str, SourceMap]:
+        result = self.client.compile(teal, source_map=source_map)
         src_map = None
         if source_map:
-            src_map = SourceMap(approval_result["sourcemap"])
-        return (b64decode(approval_result["result"]), src_map)
+            src_map = SourceMap(result["sourcemap"])
+        return (b64decode(result["result"]), result["hash"], src_map)
 
-    def compile_clear(self, source_map: bool = False) -> tuple[bytes, SourceMap]:
-        clear_result = self.client.compile(self.app.clear_program, source_map=True)
-        src_map = None
-        if source_map:
-            src_map = SourceMap(clear_result["sourcemap"])
-        return (b64decode(clear_result["result"]), src_map)
+    def build(self) -> None:
+        """
+        Wraps the Application in an AppPrecompile before calling `compile` on the Precompile which
+        recursively compiles all the dependencies (depth first). The result is then used
+        to assign the approval and clear state program binaries and src maps.
+        """
+        if self.approval_binary is not None and self.clear_binary is not None:
+            return
+
+        compiled_app = AppPrecompile(self.app)
+        compiled_app.compile(self.client)
+
+        self.approval_binary = compiled_app.approval._binary
+        self.approval_src_map = compiled_app.approval._map
+
+        self.clear_binary = compiled_app.clear._binary
+        self.clear_src_map = compiled_app.clear._map
+
+        if self.app.approval_program is not None and self.app.clear_program is not None:
+            self.approval_asserts = _gather_asserts(
+                self.app.approval_program, self.approval_src_map
+            )
+            self.clear_asserts = _gather_asserts(
+                self.app.clear_program, self.clear_src_map
+            )
 
     def create(
         self,
         sender: str = None,
         signer: TransactionSigner = None,
-        args: list[Any] = [],
+        args: list[Any] = None,
         suggested_params: transaction.SuggestedParams = None,
         on_complete: transaction.OnComplete = transaction.OnComplete.NoOpOC,
         extra_pages: int = None,
@@ -82,17 +111,16 @@ class ApplicationClient:
     ) -> tuple[int, str, str]:
         """Submits a signed ApplicationCallTransaction with application id == 0 and the schema and source from the Application passed"""
 
-        approval, approval_map = self.compile_approval()
-        self.approval_binary = approval
-        self.approval_src_map = approval_map
-
-        clear, clear_map = self.compile_clear()
-        self.clear_binary = clear
-        self.clear_src_map = clear_map
+        self.build()
+        assert self.clear_binary is not None and self.approval_binary is not None
 
         if extra_pages is None:
             extra_pages = ceil(
-                ((len(approval) + len(clear)) - APP_PAGE_MAX_SIZE) / APP_PAGE_MAX_SIZE
+                (
+                    (len(self.approval_binary) + len(self.clear_binary))
+                    - APP_PAGE_MAX_SIZE
+                )
+                / APP_PAGE_MAX_SIZE
             )
 
         sp = self.get_suggested_params(suggested_params)
@@ -100,23 +128,39 @@ class ApplicationClient:
         sender = self.get_sender(sender, signer)
 
         atc = AtomicTransactionComposer()
-        atc.add_transaction(
-            TransactionWithSigner(
-                txn=transaction.ApplicationCreateTxn(
-                    sender=sender,
-                    sp=sp,
-                    on_complete=on_complete,
-                    approval_program=approval,
-                    clear_program=clear,
-                    global_schema=self.app.app_state.schema(),
-                    local_schema=self.app.acct_state.schema(),
-                    extra_pages=extra_pages,
-                    app_args=args,
-                    **kwargs,
-                ),
-                signer=signer,
+        if self.app.on_create is not None:
+            self.add_method_call(
+                atc,
+                self.app.on_create,
+                sender=sender,
+                suggested_params=sp,
+                on_complete=on_complete,
+                approval_program=self.approval_binary,
+                clear_program=self.clear_binary,
+                global_schema=self.app.app_state.schema(),
+                local_schema=self.app.acct_state.schema(),
+                extra_pages=extra_pages,
+                app_args=args,
+                **kwargs,
+            )
+        else:
+            atc.add_transaction(
+                TransactionWithSigner(
+                    txn=transaction.ApplicationCreateTxn(
+                        sender=sender,
+                        sp=sp,
+                        on_complete=on_complete,
+                        approval_program=self.approval_binary,
+                        clear_program=self.clear_binary,
+                        global_schema=self.app.app_state.schema(),
+                        local_schema=self.app.acct_state.schema(),
+                        extra_pages=extra_pages,
+                        app_args=args,
+                        **kwargs,
+                    ),
+                    signer=signer,
+                )
             )
-        )
 
         try:
             create_result = atc.execute(self.client, 4)
@@ -141,40 +185,47 @@ class ApplicationClient:
         self,
         sender: str = None,
         signer: TransactionSigner = None,
-        args: list[Any] = [],
+        args: list[Any] = None,
         suggested_params: transaction.SuggestedParams = None,
         **kwargs,
     ) -> str:
 
         """Submits a signed ApplicationCallTransaction with OnComplete set to UpdateApplication and source from the Application passed"""
-
-        approval, approval_map = self.compile_approval()
-        self.approval_binary = approval
-        self.approval_src_map = approval_map
-
-        clear, clear_map = self.compile_clear()
-        self.clear_binary = clear
-        self.clear_src_map = clear_map
+        self.build()
 
         sp = self.get_suggested_params(suggested_params)
         signer = self.get_signer(signer)
         sender = self.get_sender(sender, signer)
 
         atc = AtomicTransactionComposer()
-        atc.add_transaction(
-            TransactionWithSigner(
-                txn=transaction.ApplicationUpdateTxn(
-                    sender=sender,
-                    sp=sp,
-                    index=self.app_id,
-                    approval_program=approval,
-                    clear_program=clear,
-                    app_args=args,
-                    **kwargs,
-                ),
-                signer=signer,
+        if self.app.on_update is not None:
+            self.add_method_call(
+                atc,
+                self.app.on_update,
+                on_complete=transaction.OnComplete.UpdateApplicationOC,
+                sender=sender,
+                suggested_params=sp,
+                index=self.app_id,
+                approval_program=self.approval_binary,
+                clear_program=self.clear_binary,
+                app_args=args,
+                **kwargs,
+            )
+        else:
+            atc.add_transaction(
+                TransactionWithSigner(
+                    txn=transaction.ApplicationUpdateTxn(
+                        sender=sender,
+                        sp=sp,
+                        index=self.app_id,
+                        approval_program=self.approval_binary,
+                        clear_program=self.clear_binary,
+                        app_args=args,
+                        **kwargs,
+                    ),
+                    signer=signer,
+                )
             )
-        )
 
         try:
             update_result = atc.execute(self.client, 4)
@@ -190,7 +241,7 @@ class ApplicationClient:
         self,
         sender: str = None,
         signer: TransactionSigner = None,
-        args: list[Any] = [],
+        args: list[Any] = None,
         suggested_params: transaction.SuggestedParams = None,
         **kwargs,
     ) -> str:
@@ -201,18 +252,31 @@ class ApplicationClient:
         sender = self.get_sender(sender, signer)
 
         atc = AtomicTransactionComposer()
-        atc.add_transaction(
-            TransactionWithSigner(
-                txn=transaction.ApplicationOptInTxn(
-                    sender=sender,
-                    sp=sp,
-                    index=self.app_id,
-                    app_args=args,
-                    **kwargs,
-                ),
+        if self.app.on_opt_in is not None:
+            self.add_method_call(
+                atc,
+                self.app.on_opt_in,
+                on_complete=transaction.OnComplete.OptInOC,
+                sender=sender,
+                suggested_params=sp,
+                index=self.app_id,
+                app_args=args,
                 signer=signer,
+                **kwargs,
+            )
+        else:
+            atc.add_transaction(
+                TransactionWithSigner(
+                    txn=transaction.ApplicationOptInTxn(
+                        sender=sender,
+                        sp=sp,
+                        index=self.app_id,
+                        app_args=args,
+                        **kwargs,
+                    ),
+                    signer=signer,
+                )
             )
-        )
 
         try:
             opt_in_result = atc.execute(self.client, 4)
@@ -228,7 +292,7 @@ class ApplicationClient:
         self,
         sender: str = None,
         signer: TransactionSigner = None,
-        args: list[Any] = [],
+        args: list[Any] = None,
         suggested_params: transaction.SuggestedParams = None,
         **kwargs,
     ) -> str:
@@ -239,18 +303,31 @@ class ApplicationClient:
         sender = self.get_sender(sender, signer)
 
         atc = AtomicTransactionComposer()
-        atc.add_transaction(
-            TransactionWithSigner(
-                txn=transaction.ApplicationCloseOutTxn(
-                    sender=sender,
-                    sp=sp,
-                    index=self.app_id,
-                    app_args=args,
-                    **kwargs,
-                ),
+        if self.app.on_close_out is not None:
+            self.add_method_call(
+                atc,
+                self.app.on_close_out,
+                on_complete=transaction.OnComplete.CloseOutOC,
+                sender=sender,
+                suggested_params=sp,
+                index=self.app_id,
+                app_args=args,
                 signer=signer,
+                **kwargs,
+            )
+        else:
+            atc.add_transaction(
+                TransactionWithSigner(
+                    txn=transaction.ApplicationCloseOutTxn(
+                        sender=sender,
+                        sp=sp,
+                        index=self.app_id,
+                        app_args=args,
+                        **kwargs,
+                    ),
+                    signer=signer,
+                )
             )
-        )
 
         try:
             close_out_result = atc.execute(self.client, 4)
@@ -266,7 +343,7 @@ class ApplicationClient:
         self,
         sender: str = None,
         signer: TransactionSigner = None,
-        args: list[Any] = [],
+        args: list[Any] = None,
         suggested_params: transaction.SuggestedParams = None,
         **kwargs,
     ) -> str:
@@ -278,18 +355,31 @@ class ApplicationClient:
         sender = self.get_sender(sender, signer)
 
         atc = AtomicTransactionComposer()
-        atc.add_transaction(
-            TransactionWithSigner(
-                txn=transaction.ApplicationClearStateTxn(
-                    sender=sender,
-                    sp=sp,
-                    index=self.app_id,
-                    app_args=args,
-                    **kwargs,
-                ),
+        if self.app.on_clear_state is not None:
+            self.add_method_call(
+                atc,
+                self.app.on_clear_state,
+                on_complete=transaction.OnComplete.ClearStateOC,
+                sender=sender,
+                suggested_params=sp,
+                index=self.app_id,
+                app_args=args,
                 signer=signer,
+                **kwargs,
+            )
+        else:
+            atc.add_transaction(
+                TransactionWithSigner(
+                    txn=transaction.ApplicationClearStateTxn(
+                        sender=sender,
+                        sp=sp,
+                        index=self.app_id,
+                        app_args=args,
+                        **kwargs,
+                    ),
+                    signer=signer,
+                )
             )
-        )
 
         clear_state_result = atc.execute(self.client, 4)
 
@@ -299,7 +389,7 @@ class ApplicationClient:
         self,
         sender: str = None,
         signer: TransactionSigner = None,
-        args: list[Any] = [],
+        args: list[Any] = None,
         suggested_params: transaction.SuggestedParams = None,
         **kwargs,
     ) -> str:
@@ -310,18 +400,31 @@ class ApplicationClient:
         sender = self.get_sender(sender, signer)
 
         atc = AtomicTransactionComposer()
-        atc.add_transaction(
-            TransactionWithSigner(
-                txn=transaction.ApplicationDeleteTxn(
-                    sender=sender,
-                    sp=sp,
-                    index=self.app_id,
-                    app_args=args,
-                    **kwargs,
-                ),
+        if self.app.on_delete:
+            self.add_method_call(
+                atc,
+                self.app.on_delete,
+                on_complete=transaction.OnComplete.DeleteApplicationOC,
+                sender=sender,
+                sp=sp,
+                index=self.app_id,
+                app_args=args,
                 signer=signer,
+                **kwargs,
+            )
+        else:
+            atc.add_transaction(
+                TransactionWithSigner(
+                    txn=transaction.ApplicationDeleteTxn(
+                        sender=sender,
+                        sp=sp,
+                        index=self.app_id,
+                        app_args=args,
+                        **kwargs,
+                    ),
+                    signer=signer,
+                )
             )
-        )
 
         try:
             delete_result = atc.execute(self.client, 4)
@@ -363,56 +466,26 @@ class ApplicationClient:
         note: bytes = None,
         lease: bytes = None,
         rekey_to: str = None,
+        atc: AtomicTransactionComposer = None,
         **kwargs,
     ) -> ABIResult:
 
         """Handles calling the application"""
 
-        sp = self.get_suggested_params(suggested_params)
-        signer = self.get_signer(signer)
-        sender = self.get_sender(sender, signer)
-
         if not isinstance(method, abi.Method):
             method = get_method_spec(method)
 
         hints = self.method_hints(method.name)
 
-        args = []
-        for method_arg in method.args:
-            name = method_arg.name
+        if atc is None:
+            atc = AtomicTransactionComposer()
 
-            if name in kwargs:
-                argument = kwargs[name]
-
-                if type(argument) is dict:
-                    if hints.structs is None or name not in hints.structs:
-                        raise Exception(f"Name {name} name in struct hints")
-
-                    argument = [
-                        argument[field_name]
-                        for field_name in hints.structs[name]["elements"]
-                    ]
-
-                args.append(argument)
-
-            elif (
-                hints.param_annotations is not None and name in hints.param_annotations
-            ):
-                annos = hints.param_annotations[name]
-                if annos.default is not None:
-                    args.append(self.resolve(DefaultArgument(annos.default)))
-            else:
-                raise Exception(f"Unspecified argument: {name}")
-
-        atc = AtomicTransactionComposer()
-
-        atc.add_method_call(
-            self.app_id,
+        atc = self.add_method_call(
+            atc,
             method,
             sender,
-            sp,
             signer,
-            method_args=args,
+            suggested_params=suggested_params,
             on_complete=on_complete,
             local_schema=local_schema,
             global_schema=global_schema,
@@ -425,11 +498,12 @@ class ApplicationClient:
             note=note,
             lease=lease,
             rekey_to=rekey_to,
+            **kwargs,
         )
 
+        # If its a read-only method, use dryrun (TODO: swap with simulate later?)
         if hints.read_only:
-            txns = atc.gather_signatures()
-            dr_req = transaction.create_dryrun(self.client, txns)
+            dr_req = transaction.create_dryrun(self.client, atc.gather_signatures())
             dr_result = self.client.dryrun(dr_req)
             method_results = self._parse_result(
                 {0: method}, dr_result["txns"], atc.tx_ids
@@ -552,18 +626,22 @@ class ApplicationClient:
                     if hints.structs is None or name not in hints.structs:
                         raise Exception(f"Name {name} name in struct hints")
 
+                    elems: list[tuple[str, str]] = cast(
+                        list[tuple[str, str]], hints.structs[name]["elements"]
+                    )
+
                     argument = [
-                        argument[field_name]
-                        for field_name in hints.structs[name]["elements"]
+                        argument[field_name] for field_name, field_type in elems
                     ]
 
                 args.append(argument)
+
             elif (
-                hints.param_annotations is not None and name in hints.param_annotations
+                hints.default_arguments is not None and name in hints.default_arguments
             ):
-                annos = hints.param_annotations[name]
-                if annos.default is not None:
-                    args.append(self.resolve(DefaultArgument(annos.default)))
+                default_arg = hints.default_arguments[name]
+                if default_arg is not None:
+                    args.append(self.resolve(default_arg))
             else:
                 raise Exception(f"Unspecified argument: {name}")
 
@@ -590,17 +668,25 @@ class ApplicationClient:
 
         return atc
 
-    def fund(self, amt: int) -> str:
-        """fund pays the app account the amount passed using the signer"""
+    def add_transaction(
+        self, atc: AtomicTransactionComposer, txn: transaction.Transaction
+    ) -> AtomicTransactionComposer:
+        atc.add_transaction(TransactionWithSigner(txn=txn, signer=self.signer))
+        return atc
+
+    def fund(self, amt: int, addr: str = None) -> str:
+        """convenience method to pay the address passed, defaults to paying the app address for this client from the current signer"""
         sender = self.get_sender()
         signer = self.get_signer()
 
         sp = self.client.suggested_params()
 
+        rcv = self.app_addr if addr is None else addr
+
         atc = AtomicTransactionComposer()
         atc.add_transaction(
             TransactionWithSigner(
-                txn=transaction.PaymentTxn(sender, sp, self.app_addr, amt),
+                txn=transaction.PaymentTxn(sender, sp, rcv, amt),
                 signer=signer,
             )
         )
@@ -642,14 +728,12 @@ class ApplicationClient:
             return to_resolve.resolve_hint()
         elif to_resolve.resolvable_class == DefaultArgumentClass.GlobalState:
             key = to_resolve.resolve_hint()
-            app_state = self.get_application_state()
-            return app_state[key]
+            app_state = self.get_application_state(raw=True)
+            return app_state[key.encode()]
         elif to_resolve.resolvable_class == DefaultArgumentClass.LocalState:
             key = to_resolve.resolve_hint()
-            acct_state = self.get_account_state(
-                self.get_sender(),
-            )
-            return acct_state[key]
+            acct_state = self.get_account_state(self.get_sender(), raw=True)
+            return acct_state[key.encode()]
         elif to_resolve.resolvable_class == DefaultArgumentClass.ABIMethod:
             method = abi.Method.undictify(to_resolve.resolve_hint())
             result = self.call(method)
@@ -665,6 +749,7 @@ class ApplicationClient:
     def get_suggested_params(
         self, sp: transaction.SuggestedParams = None
     ) -> transaction.SuggestedParams:
+
         if sp is not None:
             return sp
 
@@ -673,9 +758,11 @@ class ApplicationClient:
 
         return self.client.suggested_params()
 
-    def wrap_approval_exception(self, e: Exception, lines: int = 10) -> LogicException:
-        _, map = self.compile_approval(True)
-        return LogicException(e, self.app.approval_program, map, lines)
+    def wrap_approval_exception(self, e: Exception) -> Exception:
+        if self.app.approval_program is None or self.approval_src_map is None:
+            return e
+
+        return LogicException(e, self.app.approval_program, self.approval_src_map)
 
     def get_signer(self, signer: TransactionSigner = None) -> TransactionSigner:
         if signer is not None:
@@ -704,3 +791,31 @@ class ApplicationClient:
                 return signer.lsig.address()
 
         raise Exception("No sender provided")
+
+
+@dataclass
+class ProgramAssertion:
+    line: int
+    message: str
+
+
+def _gather_asserts(program: str, src_map: SourceMap) -> dict[int, ProgramAssertion]:
+    asserts: dict[int, ProgramAssertion] = {}
+
+    program_lines = program.split("\n")
+    for idx, line in enumerate(program_lines):
+        # Take only the first bit
+        line = line.split(" ").pop()
+        if line != "assert":
+            continue
+
+        pc = src_map.get_pcs_for_line(idx)[0]
+
+        # TODO: this will be wrong for multiline comments
+        line_before = program_lines[idx - 1]
+        if not line_before.startswith("//"):
+            continue
+
+        asserts[pc] = ProgramAssertion(idx, line_before.strip("// "))
+
+    return asserts
