@@ -13,7 +13,7 @@ from algosdk.atomic_transaction_composer import (
     LogicSigTransactionSigner,
 )
 
-from ..decorators import (
+from beaker.decorators import (
     Authorize,
     DefaultArgument,
     create,
@@ -26,11 +26,9 @@ from ..decorators import (
 )
 from beaker.sandbox import get_accounts, get_algod_client
 from beaker.application import Application, get_method_selector
-from beaker.struct import Struct
 from beaker.state import ApplicationStateValue, AccountStateValue
 from beaker.client.application_client import ApplicationClient
 from beaker.client.logic_error import LogicException
-from beaker.consts import MilliAlgo
 
 
 class App(Application):
@@ -81,14 +79,6 @@ class App(Application):
     def dummy(self, *, output: pt.abi.String):
         return output.set("deadbeef")
 
-    # class Structy(Struct):
-    #    a: pt.abi.Uint64
-    #    b: pt.abi.Uint32
-
-    # @external
-    # def structy(self, s: Structy, *, output: pt.abi.Uint64):
-    #    return output.set(s.a)
-
 
 SandboxAccounts = list[tuple[str, str, AccountTransactionSigner]]
 
@@ -205,13 +195,18 @@ def test_compile():
     client = get_algod_client()
     ac = ApplicationClient(client, app)
 
-    approval_program, approval_map = ac.compile_approval(source_map=True)
+    # TODO add precompiles
+
+    approval_program, _, approval_map = ac.compile(
+        ac.app.approval_program, source_map=True
+    )
+
     assert len(approval_program) > 0, "Should have a valid approval program"
     assert approval_program[0] == version, "First byte should be the version we set"
     assert approval_map.version == 3, "Should have valid source map with version 3"
     assert len(approval_map.pc_to_line) > 0, "Should have valid mapping"
 
-    clear_program, clear_map = ac.compile_clear(source_map=True)
+    clear_program, _, clear_map = ac.compile(ac.app.clear_program, source_map=True)
     assert len(clear_program) > 0, "Should have a valid clear program"
     assert clear_program[0] == version, "First byte should be the version we set"
     assert clear_map.version == 3, "Should have valid source map with version 3"
@@ -493,23 +488,6 @@ def test_call(sb_accts: SandboxAccounts):
     )
 
 
-# def test_call_with_struct():
-#    app = App()
-#    addr, pk, signer = sb_accts[0]
-#
-#    client = get_algod_client()
-#    ac = ApplicationClient(client, app, signer=signer)
-#    app_id, _, _ = ac.create()
-#
-#    result = ac.call(app.structy, s={'a':1, 'b':2})
-#    assert result.return_value == 1
-#    assert result.decode_error is None
-#    assert result.raw_value == (1).to_bytes(8, "big")
-#
-#    with pytest.raises(Exception):
-#        ac.call(app.structy, s=4)
-
-
 def test_add_method_call(sb_accts: SandboxAccounts):
     app = App()
 
@@ -583,7 +561,7 @@ def test_resolve(sb_accts: SandboxAccounts):
     assert ac.resolve(DefaultArgument(pt.Int(1))) == 1
     assert ac.resolve(DefaultArgument(pt.Bytes("stringy"))) == "stringy"
     assert ac.resolve(DefaultArgument(app.app_state_val_int)) == 1
-    assert ac.resolve(DefaultArgument(app.app_state_val_byte)) == "test"
+    assert ac.resolve(DefaultArgument(app.app_state_val_byte)) == b"test"
     assert ac.resolve(DefaultArgument(app.acct_state_val_int)) == 1
-    assert ac.resolve(DefaultArgument(app.acct_state_val_byte)) == "test"
+    assert ac.resolve(DefaultArgument(app.acct_state_val_byte)) == b"test"
     assert ac.resolve(DefaultArgument(app.dummy)) == "deadbeef"
